import * as react from 'react';
import react__default, { ReactElement, ComponentType, HTMLAttributes, StyleHTMLAttributes, ReactNode, ChangeEvent, ButtonHTMLAttributes } from 'react';
import * as json_schema from 'json-schema';
import { JSONSchema7 } from 'json-schema';

/** An enumeration of all the translatable strings used by `@rjsf/core` and its themes. The value of each of the
 * enumeration keys is expected to be the actual english string. Some strings contain replaceable parameter values
 * as indicated by `%1`, `%2`, etc. The number after the `%` indicates the order of the parameter. The ordering of
 * parameters is important because some languages may choose to put the second parameter before the first in its
 * translation. Also, some strings are rendered using `markdown-to-jsx` and thus support markdown and inline html.
 */
declare enum TranslatableString {
    /** Fallback title of an array item, used by ArrayField */
    ArrayItemTitle = "Item",
    /** Missing items reason, used by ArrayField */
    MissingItems = "Missing items definition",
    /** Yes label, used by BooleanField */
    YesLabel = "Yes",
    /** No label, used by BooleanField */
    NoLabel = "No",
    /** Close label, used by ErrorList */
    CloseLabel = "Close",
    /** Errors label, used by ErrorList */
    ErrorsLabel = "Errors",
    /** New additionalProperties string default value, used by ObjectField */
    NewStringDefault = "New Value",
    /** Add button title, used by AddButton */
    AddButton = "Add",
    /** Add button title, used by AddButton */
    AddItemButton = "Add Item",
    /** Copy button title, used by IconButton */
    CopyButton = "Copy",
    /** Move down button title, used by IconButton */
    MoveDownButton = "Move down",
    /** Move up button title, used by IconButton */
    MoveUpButton = "Move up",
    /** Remove button title, used by IconButton */
    RemoveButton = "Remove",
    /** Now label, used by AltDateWidget */
    NowLabel = "Now",
    /** Clear label, used by AltDateWidget */
    ClearLabel = "Clear",
    /** Aria date label, used by DateWidget */
    AriaDateLabel = "Select a date",
    /** File preview label, used by FileWidget */
    PreviewLabel = "Preview",
    /** Decrement button aria label, used by UpDownWidget */
    DecrementAriaLabel = "Decrease value by 1",
    /** Increment button aria label, used by UpDownWidget */
    IncrementAriaLabel = "Increase value by 1",
    /** Unknown field type reason, where %1 will be replaced with the type as provided by SchemaField */
    UnknownFieldType = "Unknown field type %1",
    /** Option prefix, where %1 will be replaced with the option index as provided by MultiSchemaField */
    OptionPrefix = "Option %1",
    /** Option prefix, where %1 and %2 will be replaced by the schema title and option index, respectively as provided by
     * MultiSchemaField
     */
    TitleOptionPrefix = "%1 option %2",
    /** Key label, where %1 will be replaced by the label as provided by WrapIfAdditionalTemplate */
    KeyLabel = "%1 Key",
    /** Invalid object field configuration as provided by the ObjectField */
    InvalidObjectField = "Invalid \"%1\" object field configuration: <em>%2</em>.",
    /** Unsupported field schema, used by UnsupportedField */
    UnsupportedField = "Unsupported field schema.",
    /** Unsupported field schema, where %1 will be replaced by the idSchema.$id as provided by UnsupportedField */
    UnsupportedFieldWithId = "Unsupported field schema for field <code>%1</code>.",
    /** Unsupported field schema, where %1 will be replaced by the reason string as provided by UnsupportedField */
    UnsupportedFieldWithReason = "Unsupported field schema: <em>%1</em>.",
    /** Unsupported field schema, where %1 and %2 will be replaced by the idSchema.$id and reason strings, respectively,
     * as provided by UnsupportedField
     */
    UnsupportedFieldWithIdAndReason = "Unsupported field schema for field <code>%1</code>: <em>%2</em>.",
    /** File name, type and size info, where %1, %2 and %3 will be replaced by the file name, file type and file size as
     * provided by FileWidget
     */
    FilesInfo = "<strong>%1</strong> (%2, %3 bytes)"
}

/** The representation of any generic object type, usually used as an intersection on other types to make them more
 * flexible in the properties they support (i.e. anything else)
 */
type GenericObjectType = {
    [name: string]: any;
};
/** Map the JSONSchema7 to our own type so that we can easily bump to a more recent version at some future date and only
 * have to update this one type.
 */
type StrictRJSFSchema = JSONSchema7;
/** Allow for more flexible schemas (i.e. draft-2019) than the strict JSONSchema7
 */
type RJSFSchema = StrictRJSFSchema & GenericObjectType;
/** Alias GenericObjectType as FormContextType to allow us to remap this at some future date
 */
type FormContextType = GenericObjectType;
/** Experimental feature that specifies the Array `minItems` default form state behavior
 */
type Experimental_ArrayMinItems = {
    /** Optional enumerated flag controlling how array minItems are populated, defaulting to `all`:
     * - `all`: Legacy behavior, populate minItems entries with default values initially and include an empty array when
     *        no values have been defined.
     * - `requiredOnly`: Ignore `minItems` on a field when calculating defaults unless the field is required.
     */
    populate?: 'all' | 'requiredOnly';
    /** When `formData` is provided and does not contain `minItems` worth of data, this flag (`false` by default) controls
     * whether the extra data provided by the defaults is appended onto the existing `formData` items to ensure the
     * `minItems` condition is met. When false (legacy behavior), only the `formData` provided is merged into the default
     * form state, even if there are fewer than the `minItems`. When true, the defaults are appended onto the end of the
     * `formData` until the `minItems` condition is met.
     */
    mergeExtraDefaults?: boolean;
};
/** Experimental features to specify different default form state behaviors. Currently, this affects the
 * handling of optional array fields where `minItems` is set and handling of setting defaults based on the
 * value of `emptyObjectFields`.
 */
type Experimental_DefaultFormStateBehavior = {
    /** Optional object, that controls how the default form state for arrays with `minItems` is handled. When not provided
     * it defaults to `{ populate: 'all' }`.
     */
    arrayMinItems?: Experimental_ArrayMinItems;
    /** Optional enumerated flag controlling how empty object fields are populated, defaulting to `populateAllDefaults`:
     * - `populateAllDefaults`: Legacy behavior - set default when there is a primitive value, an non-empty object field,
     *        or the field itself is required  |
     * - `populateRequiredDefaults`: Only sets default when a value is an object and its parent field is required, or it
     *        is a primitive value and it is required |
     * - `skipDefaults`: Does not set defaults                                                                                                      |
     */
    emptyObjectFields?: 'populateAllDefaults' | 'populateRequiredDefaults' | 'skipDefaults';
};
/** The interface representing a Date object that contains an optional time */
interface DateObject {
    /** The year of the Date */
    year: number;
    /** The month of the Date */
    month: number;
    /** The day of the Date */
    day: number;
    /** The optional hours for the time of a Date */
    hour?: number;
    /** The optional minutes for the time of a Date */
    minute?: number;
    /** The optional seconds for the time of a Date */
    second?: number;
}
/** Properties describing a Range specification in terms of attribute that can be added to the `HTML` `<input>` */
type RangeSpecType = {
    /** Specifies the interval between legal numbers in an input field */
    step?: number;
    /** Specifies a minimum value for an <input> element */
    min?: number;
    /** Specifies the maximum value for an <input> element */
    max?: number;
};
/** Properties describing a Range specification in terms of attribute that can be added to the `HTML` `<input>` */
type InputPropsType = Omit<RangeSpecType, 'step'> & {
    /** Specifies the type of the <input> element */
    type: string;
    /** Specifies the interval between legal numbers in an input field or "any" */
    step?: number | 'any';
    /** Specifies the `autoComplete` value for an <input> element */
    autoComplete?: HTMLInputElement['autocomplete'];
};
/** Type describing an id used for a field in the `IdSchema` */
type FieldId = {
    /** The id for a field */
    $id: string;
};
/** Type describing a recursive structure of `FieldId`s for an object with a non-empty set of keys */
type IdSchema<T = any> = FieldId & {
    [key in keyof T]?: IdSchema<T[key]>;
};
/** Type describing a name used for a field in the `PathSchema` */
type FieldPath = {
    /** The name of a field */
    $name: string;
};
/** Type describing a recursive structure of `FieldPath`s for an object with a non-empty set of keys */
type PathSchema<T = any> = FieldPath & {
    [key in keyof T]?: PathSchema<T[key]>;
};
/** The type for error produced by RJSF schema validation */
type RJSFValidationError = {
    /** Name of the error, for example, "required" or "minLength" */
    name?: string;
    /** Message, for example, "is a required property" or "should NOT be shorter than 3 characters" */
    message?: string;
    /** An object with the error params returned by ajv
     * ([see doc](https://github.com/ajv-validator/ajv/tree/6a671057ea6aae690b5967ee26a0ddf8452c6297#error-parameters)
     * for more info)
     */
    params?: any;
    /** A string in Javascript property accessor notation to the data path of the field with the error. For example,
     * `.name` or `['first-name']`
     */
    property?: string;
    /** JSON pointer to the schema of the keyword that failed validation. For example, `#/fields/firstName/required`.
     * (Note: this may sometimes be wrong due to a [bug in ajv](https://github.com/ajv-validator/ajv/issues/512))
     */
    schemaPath?: string;
    /** Full error name, for example ".name is a required property" */
    stack: string;
};
/** The type that describes an error in a field */
type FieldError = string;
/** The type that describes the list of errors for a field */
type FieldErrors = {
    /** The list of errors for the field */
    __errors?: FieldError[];
};
/** Type describing a recursive structure of `FieldErrors`s for an object with a non-empty set of keys */
type ErrorSchema<T = any> = FieldErrors & {
    [key in keyof T]?: ErrorSchema<T[key]>;
};
/** Type that describes the list of errors for a field being actively validated by a custom validator */
type FieldValidation = FieldErrors & {
    /** Function that will add a new `message` to the list of errors */
    addError: (message: string) => void;
};
/** Type describing a recursive structure of `FieldValidation`s for an object with a non-empty set of keys */
type FormValidation<T = any> = FieldValidation & {
    [key in keyof T]?: FormValidation<T[key]>;
};
/** The properties that are passed to an `ErrorListTemplate` implementation */
type ErrorListProps<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> = {
    /** The errorSchema constructed by `Form` */
    errorSchema: ErrorSchema<T>;
    /** An array of the errors */
    errors: RJSFValidationError[];
    /** The `formContext` object that was passed to `Form` */
    formContext?: F;
    /** The schema that was passed to `Form` */
    schema: S;
    /** The uiSchema that was passed to `Form` */
    uiSchema?: UiSchema<T, S, F>;
    /** The `registry` object */
    registry: Registry<T, S, F>;
};
/** The properties that are passed to an `FieldErrorTemplate` implementation */
type FieldErrorProps<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> = {
    /** The errorSchema constructed by `Form` */
    errorSchema?: ErrorSchema<T>;
    /** An array of the errors */
    errors?: Array<string | ReactElement>;
    /** The tree of unique ids for every child field */
    idSchema: IdSchema<T>;
    /** The schema that was passed to field */
    schema: S;
    /** The uiSchema that was passed to field */
    uiSchema?: UiSchema<T, S, F>;
    /** The `registry` object */
    registry: Registry<T, S, F>;
};
/** The properties that are passed to an `FieldHelpTemplate` implementation */
type FieldHelpProps<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> = {
    /** The help information to be rendered */
    help?: string | ReactElement;
    /** The tree of unique ids for every child field */
    idSchema: IdSchema<T>;
    /** The schema that was passed to field */
    schema: S;
    /** The uiSchema that was passed to field */
    uiSchema?: UiSchema<T, S, F>;
    /** Flag indicating whether there are errors associated with this field */
    hasErrors?: boolean;
    /** The `registry` object */
    registry: Registry<T, S, F>;
};
/** The set of `Fields` stored in the `Registry` */
type RegistryFieldsType<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> = {
    /** A `Field` indexed by `name` */
    [name: string]: Field<T, S, F>;
};
/** The set of `Widgets` stored in the `Registry` */
type RegistryWidgetsType<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> = {
    /** A `Widget` indexed by `name` */
    [name: string]: Widget<T, S, F>;
};
/** The set of RJSF templates that can be overridden by themes or users */
interface TemplatesType<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> {
    /** The template to use while rendering normal or fixed array fields */
    ArrayFieldTemplate: ComponentType<ArrayFieldTemplateProps<T, S, F>>;
    /** The template to use while rendering the description for an array field */
    ArrayFieldDescriptionTemplate: ComponentType<ArrayFieldDescriptionProps<T, S, F>>;
    /** The template to use while rendering an item in an array field */
    ArrayFieldItemTemplate: ComponentType<ArrayFieldTemplateItemType<T, S, F>>;
    /** The template to use while rendering the title for an array field */
    ArrayFieldTitleTemplate: ComponentType<ArrayFieldTitleProps<T, S, F>>;
    /** The template to use while rendering the standard html input */
    BaseInputTemplate: ComponentType<BaseInputTemplateProps<T, S, F>>;
    /** The template to use for rendering the description of a field */
    DescriptionFieldTemplate: ComponentType<DescriptionFieldProps<T, S, F>>;
    /** The template to use while rendering the errors for the whole form */
    ErrorListTemplate: ComponentType<ErrorListProps<T, S, F>>;
    /** The template to use while rendering the errors for a single field */
    FieldErrorTemplate: ComponentType<FieldErrorProps<T, S, F>>;
    /** The template to use while rendering the errors for a single field */
    FieldHelpTemplate: ComponentType<FieldHelpProps<T, S, F>>;
    /** The template to use while rendering a field */
    FieldTemplate: ComponentType<FieldTemplateProps<T, S, F>>;
    /** The template to use while rendering an object */
    ObjectFieldTemplate: ComponentType<ObjectFieldTemplateProps<T, S, F>>;
    /** The template to use for rendering the title of a field */
    TitleFieldTemplate: ComponentType<TitleFieldProps<T, S, F>>;
    /** The template to use for rendering information about an unsupported field type in the schema */
    UnsupportedFieldTemplate: ComponentType<UnsupportedFieldProps<T, S, F>>;
    /** The template to use for rendering a field that allows a user to add additional properties */
    WrapIfAdditionalTemplate: ComponentType<WrapIfAdditionalTemplateProps<T, S, F>>;
    /** The set of templates associated with buttons in the form */
    ButtonTemplates: {
        /** The template to use for the main `Submit` button  */
        SubmitButton: ComponentType<SubmitButtonProps<T, S, F>>;
        /** The template to use for the Add button used for AdditionalProperties and Array items */
        AddButton: ComponentType<IconButtonProps<T, S, F>>;
        /** The template to use for the Copy button used for Array items */
        CopyButton: ComponentType<IconButtonProps<T, S, F>>;
        /** The template to use for the Move Down button used for Array items */
        MoveDownButton: ComponentType<IconButtonProps<T, S, F>>;
        /** The template to use for the Move Up button used for Array items */
        MoveUpButton: ComponentType<IconButtonProps<T, S, F>>;
        /** The template to use for the Remove button used for AdditionalProperties and Array items */
        RemoveButton: ComponentType<IconButtonProps<T, S, F>>;
    };
}
/** The set of UiSchema options that can be set globally and used as fallbacks at an individual template, field or
 * widget level when no field-level value of the option is provided.
 */
type GlobalUISchemaOptions = {
    /** Flag, if set to `false`, new items cannot be added to array fields, unless overridden (defaults to true) */
    addable?: boolean;
    /** Flag, if set to `true`, array items can be copied (defaults to false) */
    copyable?: boolean;
    /** Flag, if set to `false`, array items cannot be ordered (defaults to true) */
    orderable?: boolean;
    /** Flag, if set to `false`, array items will not be removable (defaults to true) */
    removable?: boolean;
    /** Field labels are rendered by default. Labels may be omitted by setting the `label` option to `false` */
    label?: boolean;
    /** When using `additionalProperties`, key collision is prevented by appending a unique integer to the duplicate key.
     * This option allows you to change the separator between the original key name and the integer. Default is "-"
     */
    duplicateKeySuffixSeparator?: string;
};
/** The object containing the registered core, theme and custom fields and widgets as well as the root schema, form
 * context, schema utils and templates.
 */
interface Registry<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> {
    /** The set of all fields used by the `Form`. Includes fields from `core`, theme-specific fields and any custom
     * registered fields
     */
    fields: RegistryFieldsType<T, S, F>;
    /** The set of templates used by the `Form`. Includes templates from `core`, theme-specific fields and any custom
     * registered templates
     */
    templates: TemplatesType<T, S, F>;
    /** The set of all widgets used by the `Form`. Includes widgets from `core`, theme-specific widgets and any custom
     * registered widgets
     */
    widgets: RegistryWidgetsType<T, S, F>;
    /** The `formContext` object that was passed to `Form` */
    formContext: F;
    /** The root schema, as passed to the `Form`, which can contain referenced definitions */
    rootSchema: S;
    /** The current implementation of the `SchemaUtilsType` (from `@rjsf/utils`) in use by the `Form`.  Used to call any
     * of the validation-schema-based utility functions
     */
    schemaUtils: SchemaUtilsType<T, S>;
    /** The string translation function to use when displaying any of the RJSF strings in templates, fields or widgets */
    translateString: (stringKey: TranslatableString, params?: string[]) => string;
    /** The optional global UI Options that are available for all templates, fields and widgets to access */
    globalUiOptions?: GlobalUISchemaOptions;
}
/** The properties that are passed to a Field implementation */
interface FieldProps<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> extends GenericObjectType, Pick<HTMLAttributes<HTMLElement>, Exclude<keyof HTMLAttributes<HTMLElement>, 'onBlur' | 'onFocus' | 'onChange'>> {
    /** The JSON subschema object for this field */
    schema: S;
    /** The uiSchema for this field */
    uiSchema?: UiSchema<T, S, F>;
    /** The tree of unique ids for every child field */
    idSchema: IdSchema<T>;
    /** The data for this field */
    formData?: T;
    /** The tree of errors for this field and its children */
    errorSchema?: ErrorSchema<T>;
    /** The field change event handler; called with the updated form data and an optional `ErrorSchema` */
    onChange: (newFormData: T | undefined, es?: ErrorSchema<T>, id?: string) => any;
    /** The input blur event handler; call it with the field id and value */
    onBlur: (id: string, value: any) => void;
    /** The input focus event handler; call it with the field id and value */
    onFocus: (id: string, value: any) => void;
    /** The `formContext` object that you passed to `Form` */
    formContext?: F;
    /** A boolean value stating if the field should autofocus */
    autofocus?: boolean;
    /** A boolean value stating if the field is disabled */
    disabled: boolean;
    /** A boolean value stating if the field is hiding its errors */
    hideError?: boolean;
    /** A boolean value stating if the field is read-only */
    readonly: boolean;
    /** The required status of this field */
    required?: boolean;
    /** The unique name of the field, usually derived from the name of the property in the JSONSchema */
    name: string;
    /** To avoid collisions with existing ids in the DOM, it is possible to change the prefix used for ids;
     * Default is `root`
     */
    idPrefix?: string;
    /** To avoid using a path separator that is present in field names, it is possible to change the separator used for
     * ids (Default is `_`)
     */
    idSeparator?: string;
    /** An array of strings listing all generated error messages from encountered errors for this field */
    rawErrors?: string[];
    /** The `registry` object */
    registry: Registry<T, S, F>;
}
/** The definition of a React-based Field component */
type Field<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> = ComponentType<FieldProps<T, S, F>>;
/** The properties that are passed to a FieldTemplate implementation */
type FieldTemplateProps<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> = {
    /** The id of the field in the hierarchy. You can use it to render a label targeting the wrapped widget */
    id: string;
    /** A string containing the base CSS classes, merged with any custom ones defined in your uiSchema */
    classNames?: string;
    /** An object containing the style as defined in the `uiSchema` */
    style?: StyleHTMLAttributes<any>;
    /** The computed label for this field, as a string */
    label: string;
    /** A component instance rendering the field description, if one is defined (this will use any custom
     * `DescriptionField` defined)
     */
    description?: ReactElement;
    /** A string containing any `ui:description` uiSchema directive defined */
    rawDescription?: string;
    /** The field or widget component instance for this field row */
    children: ReactElement;
    /** A component instance listing any encountered errors for this field */
    errors?: ReactElement;
    /** An array of strings listing all generated error messages from encountered errors for this field */
    rawErrors?: string[];
    /** A component instance rendering any `ui:help` uiSchema directive defined */
    help?: ReactElement;
    /** A string containing any `ui:help` uiSchema directive defined. **NOTE:** `rawHelp` will be `undefined` if passed
     * `ui:help` is a React component instead of a string
     */
    rawHelp?: string;
    /** A boolean value stating if the field should be hidden */
    hidden?: boolean;
    /** A boolean value stating if the field is required */
    required?: boolean;
    /** A boolean value stating if the field is read-only */
    readonly: boolean;
    /** A boolean value stating if the field is hiding its errors */
    hideError?: boolean;
    /** A boolean value stating if the field is disabled */
    disabled: boolean;
    /** A boolean value stating if the label should be rendered or not. This is useful for nested fields in arrays where
     * you don't want to clutter the UI
     */
    displayLabel?: boolean;
    /** The schema object for this field */
    schema: S;
    /** The uiSchema object for this field */
    uiSchema?: UiSchema<T, S, F>;
    /** The `formContext` object that was passed to `Form` */
    formContext?: F;
    /** The formData for this field */
    formData?: T;
    /** The value change event handler; Can be called with a new value to change the value for this field */
    onChange: FieldProps['onChange'];
    /** The key change event handler; Called when the key associated with a field is changed for an additionalProperty */
    onKeyChange: (value: string) => () => void;
    /** The property drop/removal event handler; Called when a field is removed in an additionalProperty context */
    onDropPropertyClick: (value: string) => () => void;
    /** The `registry` object */
    registry: Registry<T, S, F>;
};
/** The properties that are passed to the `UnsupportedFieldTemplate` implementation */
type UnsupportedFieldProps<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> = {
    /** The schema object for this field */
    schema: S;
    /** The tree of unique ids for every child field */
    idSchema?: IdSchema<T>;
    /** The reason why the schema field has an unsupported type */
    reason: string;
    /** The `registry` object */
    registry: Registry<T, S, F>;
};
/** The properties that are passed to a `TitleFieldTemplate` implementation */
type TitleFieldProps<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> = {
    /** The id of the field title in the hierarchy */
    id: string;
    /** The title for the field being rendered */
    title: string;
    /** The schema object for the field being titled */
    schema: S;
    /** The uiSchema object for this title field */
    uiSchema?: UiSchema<T, S, F>;
    /** A boolean value stating if the field is required */
    required?: boolean;
    /** The `registry` object */
    registry: Registry<T, S, F>;
};
/** The properties that are passed to a `DescriptionFieldTemplate` implementation */
type DescriptionFieldProps<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> = {
    /** The id of the field description in the hierarchy */
    id: string;
    /** The schema object for the field being described */
    schema: S;
    /** The uiSchema object for this description field */
    uiSchema?: UiSchema<T, S, F>;
    /** The description of the field being rendered */
    description: string | ReactElement;
    /** The `registry` object */
    registry: Registry<T, S, F>;
};
/** The properties that are passed to a `ArrayFieldTitleTemplate` implementation */
type ArrayFieldTitleProps<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> = Omit<TitleFieldProps<T, S, F>, 'id' | 'title'> & {
    /** The title for the field being rendered */
    title?: string;
    /** The idSchema of the field in the hierarchy */
    idSchema: IdSchema<T>;
};
/** The properties that are passed to a `ArrayFieldDescriptionTemplate` implementation */
type ArrayFieldDescriptionProps<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> = Omit<DescriptionFieldProps<T, S, F>, 'id' | 'description'> & {
    /** The description of the field being rendered */
    description?: string | ReactElement;
    /** The idSchema of the field in the hierarchy */
    idSchema: IdSchema<T>;
};
/** The properties of each element in the ArrayFieldTemplateProps.items array */
type ArrayFieldTemplateItemType<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> = {
    /** The html for the item's content */
    children: ReactElement;
    /** The className string */
    className: string;
    /** A boolean value stating if the array item is disabled */
    disabled: boolean;
    /** A boolean value stating whether new items can be added to the array */
    canAdd: boolean;
    /** A boolean value stating whether the array item can be copied, assumed false if missing */
    hasCopy: boolean;
    /** A boolean value stating whether the array item can be moved down */
    hasMoveDown: boolean;
    /** A boolean value stating whether the array item can be moved up */
    hasMoveUp: boolean;
    /** A boolean value stating whether the array item can be removed */
    hasRemove: boolean;
    /** A boolean value stating whether the array item has a toolbar */
    hasToolbar: boolean;
    /** A number stating the index the array item occurs in `items` */
    index: number;
    /** A number stating the total number `items` in the array */
    totalItems: number;
    /** Returns a function that adds a new item at `index` */
    onAddIndexClick: (index: number) => (event?: any) => void;
    /** Returns a function that copies the item at `index` into the position at `index + 1` */
    onCopyIndexClick: (index: number) => (event?: any) => void;
    /** Returns a function that removes the item at `index` */
    onDropIndexClick: (index: number) => (event?: any) => void;
    /** Returns a function that swaps the items at `index` with `newIndex` */
    onReorderClick: (index: number, newIndex: number) => (event?: any) => void;
    /** A boolean value stating if the array item is read-only */
    readonly: boolean;
    /** A stable, unique key for the array item */
    key: string;
    /** The schema object for this array item */
    schema: S;
    /** The uiSchema object for this array item */
    uiSchema?: UiSchema<T, S, F>;
    /** The `registry` object */
    registry: Registry<T, S, F>;
};
/** The properties that are passed to an ArrayFieldTemplate implementation */
type ArrayFieldTemplateProps<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> = {
    /** A boolean value stating whether new elements can be added to the array */
    canAdd?: boolean;
    /** The className string */
    className?: string;
    /** A boolean value stating if the array is disabled */
    disabled?: boolean;
    /** An object containing the id for this object & ids for its properties */
    idSchema: IdSchema<T>;
    /** An array of objects representing the items in the array */
    items: ArrayFieldTemplateItemType<T, S, F>[];
    /** A function that adds a new item to the array */
    onAddClick: (event?: any) => void;
    /** A boolean value stating if the array is read-only */
    readonly?: boolean;
    /** A boolean value stating if the array is required */
    required?: boolean;
    /** A boolean value stating if the field is hiding its errors */
    hideError?: boolean;
    /** The schema object for this array */
    schema: S;
    /** The uiSchema object for this array field */
    uiSchema?: UiSchema<T, S, F>;
    /** A string value containing the title for the array */
    title: string;
    /** The `formContext` object that was passed to Form */
    formContext?: F;
    /** The formData for this array */
    formData?: T;
    /** An array of strings listing all generated error messages from encountered errors for this widget */
    rawErrors?: string[];
    /** The `registry` object */
    registry: Registry<T, S, F>;
};
/** The properties of each element in the ObjectFieldTemplateProps.properties array */
type ObjectFieldTemplatePropertyType = {
    /** The html for the property's content */
    content: ReactElement;
    /** A string representing the property name */
    name: string;
    /** A boolean value stating if the object property is disabled */
    disabled: boolean;
    /** A boolean value stating if the property is read-only */
    readonly: boolean;
    /** A boolean value stating if the property should be hidden */
    hidden: boolean;
};
/** The properties that are passed to an ObjectFieldTemplate implementation */
type ObjectFieldTemplateProps<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> = {
    /** A string value containing the title for the object */
    title: string;
    /** A string value containing the description for the object */
    description?: string;
    /** A boolean value stating if the object is disabled */
    disabled?: boolean;
    /** An array of objects representing the properties in the object */
    properties: ObjectFieldTemplatePropertyType[];
    /** Returns a function that adds a new property to the object (to be used with additionalProperties) */
    onAddClick: (schema: S) => () => void;
    /** A boolean value stating if the object is read-only */
    readonly?: boolean;
    /** A boolean value stating if the object is required */
    required?: boolean;
    /** A boolean value stating if the field is hiding its errors */
    hideError?: boolean;
    /** The schema object for this object */
    schema: S;
    /** The uiSchema object for this object field */
    uiSchema?: UiSchema<T, S, F>;
    /** An object containing the id for this object & ids for its properties */
    idSchema: IdSchema<T>;
    /** The optional validation errors in the form of an `ErrorSchema` */
    errorSchema?: ErrorSchema<T>;
    /** The form data for the object */
    formData?: T;
    /** The `formContext` object that was passed to Form */
    formContext?: F;
    /** The `registry` object */
    registry: Registry<T, S, F>;
};
/** The properties that are passed to a WrapIfAdditionalTemplate implementation */
type WrapIfAdditionalTemplateProps<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> = {
    /** The field or widget component instance for this field row */
    children: ReactNode;
} & Pick<FieldTemplateProps<T, S, F>, 'id' | 'classNames' | 'style' | 'label' | 'required' | 'readonly' | 'disabled' | 'schema' | 'uiSchema' | 'onKeyChange' | 'onDropPropertyClick' | 'registry'>;
/** The properties that are passed to a Widget implementation */
interface WidgetProps<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> extends GenericObjectType, Pick<HTMLAttributes<HTMLElement>, Exclude<keyof HTMLAttributes<HTMLElement>, 'onBlur' | 'onFocus'>> {
    /** The generated id for this widget, used to provide unique `name`s and `id`s for the HTML field elements rendered by
     * widgets
     */
    id: string;
    /** The unique name of the field, usually derived from the name of the property in the JSONSchema; Provided in support
     * of custom widgets.
     */
    name: string;
    /** The JSONSchema subschema object for this widget */
    schema: S;
    /** The uiSchema for this widget */
    uiSchema?: UiSchema<T, S, F>;
    /** The current value for this widget */
    value: any;
    /** The required status of this widget */
    required?: boolean;
    /** A boolean value stating if the widget is disabled */
    disabled?: boolean;
    /** A boolean value stating if the widget is read-only */
    readonly?: boolean;
    /** A boolean value stating if the widget is hiding its errors */
    hideError?: boolean;
    /** A boolean value stating if the widget should autofocus */
    autofocus?: boolean;
    /** The placeholder for the widget, if any */
    placeholder?: string;
    /** A map of UI Options passed as a prop to the component, including the optional `enumOptions`
     * which is a special case on top of `UIOptionsType` needed only by widgets
     */
    options: NonNullable<UIOptionsType<T, S, F>> & {
        /** The enum options list for a type that supports them */
        enumOptions?: EnumOptionsType<S>[];
    };
    /** The `formContext` object that you passed to `Form` */
    formContext?: F;
    /** The input blur event handler; call it with the widget id and value */
    onBlur: (id: string, value: any) => void;
    /** The value change event handler; call it with the new value every time it changes */
    onChange: (value: any) => void;
    /** The input focus event handler; call it with the widget id and value */
    onFocus: (id: string, value: any) => void;
    /** The computed label for this widget, as a string */
    label: string;
    /** A boolean value, if true, will cause the label to be hidden. This is useful for nested fields where you don't want
     * to clutter the UI. Customized via `label` in the `UiSchema`
     */
    hideLabel?: boolean;
    /** A boolean value stating if the widget can accept multiple values */
    multiple?: boolean;
    /** An array of strings listing all generated error messages from encountered errors for this widget */
    rawErrors?: string[];
    /** The `registry` object */
    registry: Registry<T, S, F>;
}
/** The definition of a React-based Widget component */
type Widget<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> = ComponentType<WidgetProps<T, S, F>>;
/** The properties that are passed to the BaseInputTemplate */
interface BaseInputTemplateProps<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> extends WidgetProps<T, S, F> {
    /** A `BaseInputTemplate` implements a default `onChange` handler that it passes to the HTML input component to handle
     * the `ChangeEvent`. Sometimes a widget may need to handle the `ChangeEvent` using custom logic. If that is the case,
     * that widget should provide its own handler via this prop.
     */
    onChangeOverride?: (event: ChangeEvent<HTMLInputElement>) => void;
}
/** The type that defines the props used by the Submit button */
type SubmitButtonProps<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> = {
    /** The uiSchema for this widget */
    uiSchema?: UiSchema<T, S, F>;
    /** The `registry` object */
    registry: Registry<T, S, F>;
};
/** The type that defines the props for an Icon button, extending from a basic HTML button attributes */
type IconButtonProps<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> = ButtonHTMLAttributes<HTMLButtonElement> & {
    /** An alternative specification for the type of the icon button */
    iconType?: string;
    /** The name representation or actual react element implementation for the icon */
    icon?: string | ReactElement;
    /** The uiSchema for this widget */
    uiSchema?: UiSchema<T, S, F>;
    /** The `registry` object */
    registry: Registry<T, S, F>;
};
/** The type that defines how to change the behavior of the submit button for the form */
type UISchemaSubmitButtonOptions = {
    /** The text of the submit button. Set to "Submit" by default */
    submitText?: string;
    /** Flag, if `true`, removes the submit button completely from the form */
    norender?: boolean;
    /** Any other props to be passed to the submit button itself */
    props?: GenericObjectType & {
        /** A boolean value stating if the submit button is disabled */
        disabled?: boolean;
        /** The class name for the submit button */
        className?: string;
    };
};
/** This type represents an element used to render an enum option */
type EnumOptionsType<S extends StrictRJSFSchema = RJSFSchema> = {
    /** The value for the enum option */
    value: any;
    /** The label for the enum options */
    label: string;
    /** The schema associated with the enum option when the option represents a `oneOf` or `anyOf` choice */
    schema?: S;
};
/** This type remaps the keys of `Type` to prepend `ui:` onto them. As a result it does not need to be exported */
type MakeUIType<Type> = {
    [Property in keyof Type as `ui:${string & Property}`]: Type[Property];
};
/** This type represents all the known supported options in the `ui:options` property, kept separate in order to
 * remap the keys. It also contains all the properties, optionally, of `TemplatesType` except "ButtonTemplates"
 */
type UIOptionsBaseType<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> = Partial<Omit<TemplatesType<T, S, F>, 'ButtonTemplates'>> & GlobalUISchemaOptions & {
    /** Any classnames that the user wants to be applied to a field in the ui */
    classNames?: string;
    /** Any custom style that the user wants to apply to a field in the ui, applied on the same element as classNames */
    style?: StyleHTMLAttributes<any>;
    /** We know that for title, it will be a string, if it is provided */
    title?: string;
    /** We know that for description, it will be a string, if it is provided */
    description?: string;
    /** We know that for placeholder, it will be a string, if it is provided */
    placeholder?: string;
    /** Used to add text next to a field to guide the end user in filling it in */
    help?: string;
    /** Flag, if set to `true`, will mark the field as automatically focused on a text input or textarea input */
    autofocus?: boolean;
    /** Use to mark the field as supporting auto complete on a text input or textarea input */
    autocomplete?: HTMLInputElement['autocomplete'];
    /** Flag, if set to `true`, will mark all child widgets from a given field as disabled */
    disabled?: boolean;
    /** The default value to use when an input for a field is empty */
    emptyValue?: any;
    /** Will disable any of the enum options specified in the array (by value) */
    enumDisabled?: Array<string | number | boolean>;
    /** Flag, if set to `true`, will hide the default error display for the given field AND all of its child fields in the
     * hierarchy
     */
    hideError?: boolean;
    /** Flag, if set to `true`, will mark all child widgets from a given field as read-only */
    readonly?: boolean;
    /** This property allows you to reorder the properties that are shown for a particular object */
    order?: string[];
    /** Flag, if set to `true`, will cause the `FileWidget` to show a preview (with download for non-image files) */
    filePreview?: boolean;
    /** Flag, if set to `true`, will mark a list of checkboxes as displayed all on one line instead of one per row */
    inline?: boolean;
    /** Used to change the input type (for example, `tel` or `email`) for an <input> */
    inputType?: string;
    /** Provides a means to set the initial height of a textarea widget */
    rows?: number;
    /** If submitButtonOptions is provided it should match the `UISchemaSubmitButtonOptions` type */
    submitButtonOptions?: UISchemaSubmitButtonOptions;
    /** Allows RJSF to override the default widget implementation by specifying either the name of a widget that is used
     * to look up an implementation from the `widgets` list or an actual one-off widget implementation itself
     */
    widget?: Widget<T, S, F> | string;
};
/** The type that represents the Options potentially provided by `ui:options` */
type UIOptionsType<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> = UIOptionsBaseType<T, S, F> & {
    /** Anything else will be one of these types */
    [key: string]: boolean | number | string | object | any[] | null | undefined;
};
/** Type describing the well-known properties of the `UiSchema` while also supporting all user defined properties,
 * starting with `ui:`.
 */
type UiSchema<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> = GenericObjectType & MakeUIType<UIOptionsBaseType<T, S, F>> & {
    /** The set of Globally relevant UI Schema options that are read from the root-level UiSchema and stored in the
     * Registry for use everywhere.
     */
    'ui:globalOptions'?: GlobalUISchemaOptions;
    /** Allows the form to generate a unique prefix for the `Form`'s root prefix  */
    'ui:rootFieldId'?: string;
    /** Allows RJSF to override the default field implementation by specifying either the name of a field that is used
     * to look up an implementation from the `fields` list or an actual one-off `Field` component implementation itself
     */
    'ui:field'?: Field<T, S, F> | string;
    /** By default, any field that is rendered for an `anyOf`/`oneOf` schema will be wrapped inside the `AnyOfField` or
     * `OneOfField` component. This default behavior may be undesirable if your custom field already handles behavior
     * related to choosing one or more subschemas contained in the `anyOf`/`oneOf` schema.
     * By providing a `true` value for this flag in association with a custom `ui:field`, the wrapped components will be
     * omitted, so just one instance of the custom field will be rendered. If the flag is omitted or set to `false`,
     * your custom field will be wrapped by `AnyOfField`/`OneOfField`.
     */
    'ui:fieldReplacesAnyOrOneOf'?: boolean;
    /** An object that contains all the potential UI options in a single object */
    'ui:options'?: UIOptionsType<T, S, F>;
};
/** A `CustomValidator` function takes in a `formData`, `errors` and `uiSchema` objects and returns the given `errors`
 * object back, while potentially adding additional messages to the `errors`
 */
type CustomValidator<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> = (formData: T | undefined, errors: FormValidation<T>, uiSchema?: UiSchema<T, S, F>) => FormValidation<T>;
/** An `ErrorTransformer` function will take in a list of `errors` & a `uiSchema` and potentially return a
 * transformation of those errors in what ever way it deems necessary
 */
type ErrorTransformer<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> = (errors: RJSFValidationError[], uiSchema?: UiSchema<T, S, F>) => RJSFValidationError[];
/** The type that describes the data that is returned from the `ValidatorType.validateFormData()` function */
type ValidationData<T> = {
    /** The validation errors as a list of `RJSFValidationError` objects */
    errors: RJSFValidationError[];
    /** The validation errors in the form of an `ErrorSchema` */
    errorSchema: ErrorSchema<T>;
};
/** The interface that describes the validation functions that are provided by a Validator implementation used by the
 * schema utilities.
 */
interface ValidatorType<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> {
    /** This function processes the `formData` with an optional user contributed `customValidate` function, which receives
     * the form data and a `errorHandler` function that will be used to add custom validation errors for each field. Also
     * supports a `transformErrors` function that will take the raw AJV validation errors, prior to custom validation and
     * transform them in what ever way it chooses.
     *
     * @param formData - The form data to validate
     * @param schema - The schema against which to validate the form data
     * @param [customValidate] - An optional function that is used to perform custom validation
     * @param [transformErrors] - An optional function that is used to transform errors after AJV validation
     * @param [uiSchema] - An optional uiSchema that is passed to `transformErrors` and `customValidate`
     */
    validateFormData(formData: T | undefined, schema: S, customValidate?: CustomValidator<T, S, F>, transformErrors?: ErrorTransformer<T, S, F>, uiSchema?: UiSchema<T, S, F>): ValidationData<T>;
    /** Converts an `errorSchema` into a list of `RJSFValidationErrors`
     *
     * @param errorSchema - The `ErrorSchema` instance to convert
     * @param [fieldPath=[]] - The current field path, defaults to [] if not specified
     * @deprecated - Use the `toErrorList()` function provided by `@rjsf/utils` instead. This function will be removed in
     *        the next major release.
     */
    toErrorList(errorSchema?: ErrorSchema<T>, fieldPath?: string[]): RJSFValidationError[];
    /** Validates data against a schema, returning true if the data is valid, or
     * false otherwise. If the schema is invalid, then this function will return
     * false.
     *
     * @param schema - The schema against which to validate the form data   * @param schema
     * @param formData - The form data to validate
     * @param rootSchema - The root schema used to provide $ref resolutions
     */
    isValid(schema: S, formData: T | undefined, rootSchema: S): boolean;
    /** Runs the pure validation of the `schema` and `formData` without any of the RJSF functionality. Provided for use
     * by the playground. Returns the `errors` from the validation
     *
     * @param schema - The schema against which to validate the form data
     * @param formData - The form data to validate
     */
    rawValidation<Result = any>(schema: S, formData?: T): {
        errors?: Result[];
        validationError?: Error;
    };
}
/** The `SchemaUtilsType` interface provides a wrapper around the publicly exported APIs in the `@rjsf/utils/schema`
 * directory such that one does not have to explicitly pass the `validator` or `rootSchema` to each method. Since both
 * the `validator` and `rootSchema` generally does not change across a `Form`, this allows for providing a simplified
 * set of APIs to the `@rjsf/core` components and the various themes as well.
 */
interface SchemaUtilsType<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any> {
    /** Returns the `ValidatorType` in the `SchemaUtilsType`
     *
     * @returns - The `ValidatorType`
     */
    getValidator(): ValidatorType<T, S, F>;
    /** Determines whether either the `validator` and `rootSchema` differ from the ones associated with this instance of
     * the `SchemaUtilsType`. If either `validator` or `rootSchema` are falsy, then return false to prevent the creation
     * of a new `SchemaUtilsType` with incomplete properties.
     *
     * @param validator - An implementation of the `ValidatorType` interface that will be compared against the current one
     * @param rootSchema - The root schema that will be compared against the current one
     * @param [experimental_defaultFormStateBehavior] - Optional configuration object, if provided, allows users to override default form state behavior
     * @returns - True if the `SchemaUtilsType` differs from the given `validator` or `rootSchema`
     */
    doesSchemaUtilsDiffer(validator: ValidatorType<T, S, F>, rootSchema: S, experimental_defaultFormStateBehavior?: Experimental_DefaultFormStateBehavior): boolean;
    /** Returns the superset of `formData` that includes the given set updated to include any missing fields that have
     * computed to have defaults provided in the `schema`.
     *
     * @param schema - The schema for which the default state is desired
     * @param [formData] - The current formData, if any, onto which to provide any missing defaults
     * @param [includeUndefinedValues=false] - Optional flag, if true, cause undefined values to be added as defaults.
     *          If "excludeObjectChildren", cause undefined values for this object and pass `includeUndefinedValues` as
     *          false when computing defaults for any nested object properties.
     * @returns - The resulting `formData` with all the defaults provided
     */
    getDefaultFormState(schema: S, formData?: T, includeUndefinedValues?: boolean | 'excludeObjectChildren'): T | T[] | undefined;
    /** Determines whether the combination of `schema` and `uiSchema` properties indicates that the label for the `schema`
     * should be displayed in a UI.
     *
     * @param schema - The schema for which the display label flag is desired
     * @param [uiSchema] - The UI schema from which to derive potentially displayable information
     * @param [globalOptions={}] - The Global UI Schema from which to get any fallback `xxx` options
     * @returns - True if the label should be displayed or false if it should not
     */
    getDisplayLabel(schema: S, uiSchema?: UiSchema<T, S, F>, globalOptions?: GlobalUISchemaOptions): boolean;
    /** Determines which of the given `options` provided most closely matches the `formData`.
     * Returns the index of the option that is valid and is the closest match, or 0 if there is no match.
     *
     * The closest match is determined using the number of matching properties, and more heavily favors options with
     * matching readOnly, default, or const values.
     *
     * @param formData - The form data associated with the schema
     * @param options - The list of options that can be selected from
     * @param [selectedOption] - The index of the currently selected option, defaulted to -1 if not specified
     * @param [discriminatorField] - The optional name of the field within the options object whose value is used to
     *          determine which option is selected
     * @returns - The index of the option that is the closest match to the `formData` or the `selectedOption` if no match
     */
    getClosestMatchingOption(formData: T | undefined, options: S[], selectedOption?: number, discriminatorField?: string): number;
    /** Given the `formData` and list of `options`, attempts to find the index of the first option that matches the data.
     * Always returns the first option if there is nothing that matches.
     *
     * @param formData - The current formData, if any, used to figure out a match
     * @param options - The list of options to find a matching options from
     * @param [discriminatorField] - The optional name of the field within the options object whose value is used to
     *          determine which option is selected
     * @returns - The firstindex of the matched option or 0 if none is available
     */
    getFirstMatchingOption(formData: T | undefined, options: S[], discriminatorField?: string): number;
    /** Given the `formData` and list of `options`, attempts to find the index of the option that best matches the data.
     * Deprecated, use `getFirstMatchingOption()` instead.
     *
     * @param formData - The current formData, if any, onto which to provide any missing defaults
     * @param options - The list of options to find a matching options from
     * @param [discriminatorField] - The optional name of the field within the options object whose value is used to
     *          determine which option is selected
     * @returns - The index of the matched option or 0 if none is available
     * @deprecated
     */
    getMatchingOption(formData: T | undefined, options: S[], discriminatorField?: string): number;
    /** Checks to see if the `schema` and `uiSchema` combination represents an array of files
     *
     * @param schema - The schema for which check for array of files flag is desired
     * @param [uiSchema] - The UI schema from which to check the widget
     * @returns - True if schema/uiSchema contains an array of files, otherwise false
     */
    isFilesArray(schema: S, uiSchema?: UiSchema<T, S, F>): boolean;
    /** Checks to see if the `schema` combination represents a multi-select
     *
     * @param schema - The schema for which check for a multi-select flag is desired
     * @returns - True if schema contains a multi-select, otherwise false
     */
    isMultiSelect(schema: S): boolean;
    /** Checks to see if the `schema` combination represents a select
     *
     * @param schema - The schema for which check for a select flag is desired
     * @returns - True if schema contains a select, otherwise false
     */
    isSelect(schema: S): boolean;
    /** Merges the errors in `additionalErrorSchema` into the existing `validationData` by combining the hierarchies in
     * the two `ErrorSchema`s and then appending the error list from the `additionalErrorSchema` obtained by calling
     * `validator.toErrorList()` onto the `errors` in the `validationData`. If no `additionalErrorSchema` is passed, then
     * `validationData` is returned.
     *
     * @param validationData - The current `ValidationData` into which to merge the additional errors
     * @param [additionalErrorSchema] - The additional set of errors
     * @returns - The `validationData` with the additional errors from `additionalErrorSchema` merged into it, if provided
     * @deprecated - Use the `validationDataMerge()` function exported from `@rjsf/utils` instead. This function will be
     *        removed in the next major release.
     */
    mergeValidationData(validationData: ValidationData<T>, additionalErrorSchema?: ErrorSchema<T>): ValidationData<T>;
    /** Retrieves an expanded schema that has had all of its conditions, additional properties, references and
     * dependencies resolved and merged into the `schema` given a `rawFormData` that is used to do the potentially
     * recursive resolution.
     *
     * @param schema - The schema for which retrieving a schema is desired
     * @param [formData] - The current formData, if any, to assist retrieving a schema
     * @returns - The schema having its conditions, additional properties, references and dependencies resolved
     */
    retrieveSchema(schema: S, formData?: T): S;
    /** Sanitize the `data` associated with the `oldSchema` so it is considered appropriate for the `newSchema`. If the
     * new schema does not contain any properties, then `undefined` is returned to clear all the form data. Due to the
     * nature of schemas, this sanitization happens recursively for nested objects of data. Also, any properties in the
     * old schema that are non-existent in the new schema are set to `undefined`.
     *
     * @param [newSchema] - The new schema for which the data is being sanitized
     * @param [oldSchema] - The old schema from which the data originated
     * @param [data={}] - The form data associated with the schema, defaulting to an empty object when undefined
     * @returns - The new form data, with all of the fields uniquely associated with the old schema set
     *      to `undefined`. Will return `undefined` if the new schema is not an object containing properties.
     */
    sanitizeDataForNewSchema(newSchema?: S, oldSchema?: S, data?: any): T;
    /** Generates an `IdSchema` object for the `schema`, recursively
     *
     * @param schema - The schema for which the display label flag is desired
     * @param [id] - The base id for the schema
     * @param [formData] - The current formData, if any, onto which to provide any missing defaults
     * @param [idPrefix='root'] - The prefix to use for the id
     * @param [idSeparator='_'] - The separator to use for the path segments in the id
     * @returns - The `IdSchema` object for the `schema`
     */
    toIdSchema(schema: S, id?: string, formData?: T, idPrefix?: string, idSeparator?: string): IdSchema<T>;
    /** Generates an `PathSchema` object for the `schema`, recursively
     *
     * @param schema - The schema for which the display label flag is desired
     * @param [name] - The base name for the schema
     * @param [formData] - The current formData, if any, onto which to provide any missing defaults
     * @returns - The `PathSchema` object for the `schema`
     */
    toPathSchema(schema: S, name?: string, formData?: T): PathSchema<T>;
}

/** Checks the schema to see if it is allowing additional items, by verifying that `schema.additionalItems` is an
 * object. The user is warned in the console if `schema.additionalItems` has the value `true`.
 *
 * @param schema - The schema object to check
 * @returns - True if additional items is allowed, otherwise false
 */
declare function allowAdditionalItems<S extends StrictRJSFSchema = RJSFSchema>(schema: S): boolean;

/** Attempts to convert the string into a number. If an empty string is provided, then `undefined` is returned. If a
 * `null` is provided, it is returned. If the string ends in a `.` then the string is returned because the user may be
 * in the middle of typing a float number. If a number ends in a pattern like `.0`, `.20`, `.030`, string is returned
 * because the user may be typing number that will end in a non-zero digit. Otherwise, the string is wrapped by
 * `Number()` and if that result is not `NaN`, that number will be returned, otherwise the string `value` will be.
 *
 * @param value - The string or null value to convert to a number
 * @returns - The `value` converted to a number when appropriate, otherwise the `value`
 */
declare function asNumber(value: string | null): string | number | null | undefined;

/** Checks whether the field described by `schema`, having the `uiSchema` and `formData` supports expanding. The UI for
 * the field can expand if it has additional properties, is not forced as non-expandable by the `uiSchema` and the
 * `formData` object doesn't already have `schema.maxProperties` elements.
 *
 * @param schema - The schema for the field that is being checked
 * @param [uiSchema={}] - The uiSchema for the field
 * @param [formData] - The formData for the field
 * @returns - True if the schema element has additionalProperties, is expandable, and not at the maxProperties limit
 */
declare function canExpand<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any>(schema: RJSFSchema, uiSchema?: UiSchema<T, S, F>, formData?: T): boolean;

/** Given a `formData` object, recursively creates a `FormValidation` error handling structure around it
 *
 * @param formData - The form data around which the error handler is created
 * @returns - A `FormValidation` object based on the `formData` structure
 */
declare function createErrorHandler<T = any>(formData: T): FormValidation<T>;

/** Creates a `SchemaUtilsType` interface that is based around the given `validator` and `rootSchema` parameters. The
 * resulting interface implementation will forward the `validator` and `rootSchema` to all the wrapped APIs.
 *
 * @param validator - an implementation of the `ValidatorType` interface that will be forwarded to all the APIs
 * @param rootSchema - The root schema that will be forwarded to all the APIs
 * @param [experimental_defaultFormStateBehavior] Optional configuration object, if provided, allows users to override default form state behavior
 * @returns - An implementation of a `SchemaUtilsType` interface
 */
declare function createSchemaUtils<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any>(validator: ValidatorType<T, S, F>, rootSchema: S, experimental_defaultFormStateBehavior?: {}): SchemaUtilsType<T, S, F>;

/** Given the `FileReader.readAsDataURL()` based `dataURI` extracts that data into an actual Blob along with the name
 * of that Blob if provided in the URL. If no name is provided, then the name falls back to `unknown`.
 *
 * @param dataURI - The `DataUrl` potentially containing name and raw data to be converted to a Blob
 * @returns - an object containing a Blob and its name, extracted from the URI
 */
declare function dataURItoBlob(dataURI: string): {
    blob: Blob;
    name: string;
} | {
    blob: {
        size: number;
        type: string;
    };
    name: string;
};

/** Implements a deep equals using the `lodash.isEqualWith` function, that provides a customized comparator that
 * assumes all functions are equivalent.
 *
 * @param a - The first element to compare
 * @param b - The second element to compare
 * @returns - True if the `a` and `b` are deeply equal, false otherwise
 */
declare function deepEquals(a: any, b: any): boolean;

/** Translates a `TranslatableString` value `stringToTranslate` into english. When a `params` array is provided, each
 * value in the array is used to replace any of the replaceable parameters in the `stringToTranslate` using the `%1`,
 * `%2`, etc. replacement specifiers.
 *
 * @param stringToTranslate - The `TranslatableString` value to convert to english
 * @param params - The optional list of replaceable parameter values to substitute into the english string
 * @returns - The `stringToTranslate` itself with any replaceable parameter values substituted
 */
declare function englishStringTranslator(stringToTranslate: TranslatableString, params?: string[]): string;

/** Removes the enum option value at the `valueIndex` from the currently `selected` (list of) value(s). If `selected` is
 * a list, then that list is updated to remove the enum option value with the `valueIndex` in `allEnumOptions`. If it is
 * a single value, then if the enum option value with the `valueIndex` in `allEnumOptions` matches `selected`, undefined
 * is returned, otherwise the `selected` value is returned.
 *
 * @param valueIndex - The index of the value to be removed from the selected list or single value
 * @param selected - The current (list of) selected value(s)
 * @param [allEnumOptions=[]] - The list of all the known enumOptions
 * @returns - The updated `selected` with the enum option value at `valueIndex` in `allEnumOptions` removed from it,
 *        unless `selected` is a single value. In that case, if the `valueIndex` value matches `selected`, returns
 *        undefined, otherwise `selected`.
 */
declare function enumOptionsDeselectValue<S extends StrictRJSFSchema = RJSFSchema>(valueIndex: string | number, selected?: EnumOptionsType<S>['value'] | EnumOptionsType<S>['value'][], allEnumOptions?: EnumOptionsType<S>[]): EnumOptionsType<S>['value'] | EnumOptionsType<S>['value'][] | undefined;

/** Returns the index(es) of the options in `allEnumOptions` whose value(s) match the ones in `value`. All the
 * `enumOptions` are filtered based on whether they are a "selected" `value` and the index of each selected one is then
 * stored in an array. If `multiple` is true, that array is returned, otherwise the first element in the array is
 * returned.
 *
 * @param value - The single value or list of values for which indexes are desired
 * @param [allEnumOptions=[]] - The list of all the known enumOptions
 * @param [multiple=false] - Optional flag, if true will return a list of index, otherwise a single one
 * @returns - A single string index for the first `value` in `allEnumOptions`, if not `multiple`. Otherwise, the list
 *        of indexes for (each of) the value(s) in `value`.
 */
declare function enumOptionsIndexForValue<S extends StrictRJSFSchema = RJSFSchema>(value: EnumOptionsType<S>['value'] | EnumOptionsType<S>['value'][], allEnumOptions?: EnumOptionsType<S>[], multiple?: boolean): string | string[] | undefined;

/** Determines whether the given `value` is (one of) the `selected` value(s).
 *
 * @param value - The value being checked to see if it is selected
 * @param selected - The current selected value or list of values
 * @returns - true if the `value` is one of the `selected` ones, false otherwise
 */
declare function enumOptionsIsSelected<S extends StrictRJSFSchema = RJSFSchema>(value: EnumOptionsType<S>['value'], selected: EnumOptionsType<S>['value'] | EnumOptionsType<S>['value'][]): boolean;

/** Add the enum option value at the `valueIndex` to the list of `selected` values in the proper order as defined by
 * `allEnumOptions`
 *
 * @param valueIndex - The index of the value that should be selected
 * @param selected - The current list of selected values
 * @param [allEnumOptions=[]] - The list of all the known enumOptions
 * @returns - The updated list of selected enum values with enum value at the `valueIndex` added to it
 */
declare function enumOptionsSelectValue<S extends StrictRJSFSchema = RJSFSchema>(valueIndex: string | number, selected: EnumOptionsType<S>['value'][], allEnumOptions?: EnumOptionsType<S>[]): any[];

/** Returns the value(s) from `allEnumOptions` at the index(es) provided by `valueIndex`. If `valueIndex` is not an
 * array AND the index is not valid for `allEnumOptions`, `emptyValue` is returned. If `valueIndex` is an array, AND it
 * contains an invalid index, the returned array will have the resulting undefined values filtered out, leaving only
 * valid values or in the worst case, an empty array.
 *
 * @param valueIndex - The index(es) of the value(s) that should be returned
 * @param [allEnumOptions=[]] - The list of all the known enumOptions
 * @param [emptyValue] - The value to return when the non-array `valueIndex` does not refer to a real option
 * @returns - The single or list of values specified by the single or list of indexes if they are valid. Otherwise,
 *        `emptyValue` or an empty list.
 */
declare function enumOptionsValueForIndex<S extends StrictRJSFSchema = RJSFSchema>(valueIndex: string | number | Array<string | number>, allEnumOptions?: EnumOptionsType<S>[], emptyValue?: EnumOptionsType<S>['value']): EnumOptionsType<S>['value'] | EnumOptionsType<S>['value'][] | undefined;

/** The `ErrorSchemaBuilder<T>` is used to build an `ErrorSchema<T>` since the definition of the `ErrorSchema` type is
 * designed for reading information rather than writing it. Use this class to add, replace or clear errors in an error
 * schema by using either dotted path or an array of path names. Once you are done building the `ErrorSchema`, you can
 * get the result and/or reset all the errors back to an initial set and start again.
 */
declare class ErrorSchemaBuilder<T = any> {
    /** The error schema being built
     *
     * @private
     */
    private errorSchema;
    /** Construct an `ErrorSchemaBuilder` with an optional initial set of errors in an `ErrorSchema`.
     *
     * @param [initialSchema] - The optional set of initial errors, that will be cloned into the class
     */
    constructor(initialSchema?: ErrorSchema<T>);
    /** Returns the `ErrorSchema` that has been updated by the methods of the `ErrorSchemaBuilder`
     */
    get ErrorSchema(): ErrorSchema<T>;
    /** Will get an existing `ErrorSchema` at the specified `pathOfError` or create and return one.
     *
     * @param [pathOfError] - The optional path into the `ErrorSchema` at which to add the error(s)
     * @returns - The error block for the given `pathOfError` or the root if not provided
     * @private
     */
    private getOrCreateErrorBlock;
    /** Resets all errors in the `ErrorSchemaBuilder` back to the `initialSchema` if provided, otherwise an empty set.
     *
     * @param [initialSchema] - The optional set of initial errors, that will be cloned into the class
     * @returns - The `ErrorSchemaBuilder` object for chaining purposes
     */
    resetAllErrors(initialSchema?: ErrorSchema<T>): this;
    /** Adds the `errorOrList` to the list of errors in the `ErrorSchema` at either the root level or the location within
     * the schema described by the `pathOfError`. For more information about how to specify the path see the
     * [eslint lodash plugin docs](https://github.com/wix/eslint-plugin-lodash/blob/master/docs/rules/path-style.md).
     *
     * @param errorOrList - The error or list of errors to add into the `ErrorSchema`
     * @param [pathOfError] - The optional path into the `ErrorSchema` at which to add the error(s)
     * @returns - The `ErrorSchemaBuilder` object for chaining purposes
     */
    addErrors(errorOrList: string | string[], pathOfError?: string | string[]): this;
    /** Sets/replaces the `errorOrList` as the error(s) in the `ErrorSchema` at either the root level or the location
     * within the schema described by the `pathOfError`. For more information about how to specify the path see the
     * [eslint lodash plugin docs](https://github.com/wix/eslint-plugin-lodash/blob/master/docs/rules/path-style.md).
     *
     * @param errorOrList - The error or list of errors to set into the `ErrorSchema`
     * @param [pathOfError] - The optional path into the `ErrorSchema` at which to set the error(s)
     * @returns - The `ErrorSchemaBuilder` object for chaining purposes
     */
    setErrors(errorOrList: string | string[], pathOfError?: string | string[]): this;
    /** Clears the error(s) in the `ErrorSchema` at either the root level or the location within the schema described by
     * the `pathOfError`. For more information about how to specify the path see the
     * [eslint lodash plugin docs](https://github.com/wix/eslint-plugin-lodash/blob/master/docs/rules/path-style.md).
     *
     * @param [pathOfError] - The optional path into the `ErrorSchema` at which to clear the error(s)
     * @returns - The `ErrorSchemaBuilder` object for chaining purposes
     */
    clearErrors(pathOfError?: string | string[]): this;
}

/** Given the name of a `$ref` from within a schema, using the `rootSchema`, look up and return the sub-schema using the
 * path provided by that reference. If `#` is not the first character of the reference, or the path does not exist in
 * the schema, then throw an Error. Otherwise return the sub-schema. Also deals with nested `$ref`s in the sub-schema.
 *
 * @param $ref - The ref string for which the schema definition is desired
 * @param [rootSchema={}] - The root schema in which to search for the definition
 * @returns - The sub-schema within the `rootSchema` which matches the `$ref` if it exists
 * @throws - Error indicating that no schema for that reference exists
 */
declare function findSchemaDefinition<S extends StrictRJSFSchema = RJSFSchema>($ref?: string, rootSchema?: S): S;

/** Returns the `discriminator.propertyName` when defined in the `schema` if it is a string. A warning is generated when
 * it is not a string. Returns `undefined` when a valid discriminator is not present.
 *
 * @param schema - The schema from which the discriminator is potentially obtained
 * @returns - The `discriminator.propertyName` if it exists in the schema, otherwise `undefined`
 */
declare function getDiscriminatorFieldFromSchema<S extends StrictRJSFSchema = RJSFSchema>(schema: S): string | undefined;

/** Using the `schema`, `defaultType` and `options`, extract out the props for the <input> element that make sense.
 *
 * @param schema - The schema for the field provided by the widget
 * @param [defaultType] - The default type, if any, for the field provided by the widget
 * @param [options={}] - The UI Options for the field provided by the widget
 * @param [autoDefaultStepAny=true] - Determines whether to auto-default step=any when the type is number and no step
 * @returns - The extracted `InputPropsType` object
 */
declare function getInputProps<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any>(schema: RJSFSchema, defaultType?: string, options?: UIOptionsType<T, S, F>, autoDefaultStepAny?: boolean): InputPropsType;

/** Gets the type of a given `schema`. If the type is not explicitly defined, then an attempt is made to infer it from
 * other elements of the schema as follows:
 * - schema.const: Returns the `guessType()` of that value
 * - schema.enum: Returns `string`
 * - schema.properties: Returns `object`
 * - schema.additionalProperties: Returns `object`
 * - type is an array with a length of 2 and one type is 'null': Returns the other type
 *
 * @param schema - The schema for which to get the type
 * @returns - The type of the schema
 */
declare function getSchemaType<S extends StrictRJSFSchema = RJSFSchema>(schema: S): string | string[] | undefined;

/** Extracts any `ui:submitButtonOptions` from the `uiSchema` and merges them onto the `DEFAULT_OPTIONS`
 *
 * @param [uiSchema={}] - the UI Schema from which to extract submit button props
 * @returns - The merging of the `DEFAULT_OPTIONS` with any custom ones
 */
declare function getSubmitButtonOptions<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any>(uiSchema?: UiSchema<T, S, F>): UISchemaSubmitButtonOptions;

/** Returns the template with the given `name` from either the `uiSchema` if it is defined or from the `registry`
 * otherwise. NOTE, since `ButtonTemplates` are not overridden in `uiSchema` only those in the `registry` are returned.
 *
 * @param name - The name of the template to fetch, restricted to the keys of `TemplatesType`
 * @param registry - The `Registry` from which to read the template
 * @param [uiOptions={}] - The `UIOptionsType` from which to read an alternate template
 * @returns - The template from either the `uiSchema` or `registry` for the `name`
 */
declare function getTemplate<Name extends keyof TemplatesType<T, S, F>, T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any>(name: Name, registry: Registry<T, S, F>, uiOptions?: UIOptionsType<T, S, F>): TemplatesType<T, S, F>[Name];

/** Get all passed options from ui:options, and ui:<optionName>, returning them in an object with the `ui:`
 * stripped off. Any `globalOptions` will always be returned, unless they are overridden by options in the `uiSchema`.
 *
 * @param [uiSchema={}] - The UI Schema from which to get any `ui:xxx` options
 * @param [globalOptions={}] - The optional Global UI Schema from which to get any fallback `xxx` options
 * @returns - An object containing all the `ui:xxx` options with the `ui:` stripped off along with all `globalOptions`
 */
declare function getUiOptions<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any>(uiSchema?: UiSchema<T, S, F>, globalOptions?: GlobalUISchemaOptions): UIOptionsType<T, S, F>;

/** Given a schema representing a field to render and either the name or actual `Widget` implementation, returns the
 * React component that is used to render the widget. If the `widget` is already a React component, then it is wrapped
 * with a `MergedWidget`. Otherwise an attempt is made to look up the widget inside of the `registeredWidgets` map based
 * on the schema type and `widget` name. If no widget component can be found an `Error` is thrown.
 *
 * @param schema - The schema for the field
 * @param [widget] - Either the name of the widget OR a `Widget` implementation to use
 * @param [registeredWidgets={}] - A registry of widget name to `Widget` implementation
 * @returns - The `Widget` component to use
 * @throws - An error if there is no `Widget` component that can be returned
 */
declare function getWidget<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any>(schema: RJSFSchema, widget?: Widget<T, S, F> | string, registeredWidgets?: RegistryWidgetsType<T, S, F>): Widget<T, S, F>;

/** Given a specific `value` attempts to guess the type of a schema element. In the case where we have to implicitly
 *  create a schema, it is useful to know what type to use based on the data we are defining.
 *
 * @param value - The value from which to guess the type
 * @returns - The best guess for the object type
 */
declare function guessType(value: any): "array" | "string" | "null" | "boolean" | "number" | "object";

/** Stringifies the schema and returns the hash of the resulting string. Sorts schema fields
 * in consistent order before stringify to prevent different hash ids for the same schema.
 *
 * @param schema - The schema for which the hash is desired
 * @returns - The string obtained from the hash of the stringified schema
 */
declare function hashForSchema<S extends StrictRJSFSchema = RJSFSchema>(schema: S): string;

/** Detects whether the `widget` exists for the `schema` with the associated `registryWidgets` and returns true if it
 * does, or false if it doesn't.
 *
 * @param schema - The schema for the field
 * @param widget - Either the name of the widget OR a `Widget` implementation to use
 * @param [registeredWidgets={}] - A registry of widget name to `Widget` implementation
 * @returns - True if the widget exists, false otherwise
 */
declare function hasWidget<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any>(schema: RJSFSchema, widget: Widget<T, S, F> | string, registeredWidgets?: RegistryWidgetsType<T, S, F>): boolean;

/** Return a consistent `id` for the field description element
 *
 * @param id - Either simple string id or an IdSchema from which to extract it
 * @returns - The consistent id for the field description element from the given `id`
 */
declare function descriptionId<T = any>(id: IdSchema<T> | string): string;
/** Return a consistent `id` for the field error element
 *
 * @param id - Either simple string id or an IdSchema from which to extract it
 * @returns - The consistent id for the field error element from the given `id`
 */
declare function errorId<T = any>(id: IdSchema<T> | string): string;
/** Return a consistent `id` for the field examples element
 *
 * @param id - Either simple string id or an IdSchema from which to extract it
 * @returns - The consistent id for the field examples element from the given `id`
 */
declare function examplesId<T = any>(id: IdSchema<T> | string): string;
/** Return a consistent `id` for the field help element
 *
 * @param id - Either simple string id or an IdSchema from which to extract it
 * @returns - The consistent id for the field help element from the given `id`
 */
declare function helpId<T = any>(id: IdSchema<T> | string): string;
/** Return a consistent `id` for the field title element
 *
 * @param id - Either simple string id or an IdSchema from which to extract it
 * @returns - The consistent id for the field title element from the given `id`
 */
declare function titleId<T = any>(id: IdSchema<T> | string): string;
/** Return a list of element ids that contain additional information about the field that can be used to as the aria
 * description of the field. This is correctly omitting `titleId` which would be "labeling" rather than "describing" the
 * element.
 *
 * @param id - Either simple string id or an IdSchema from which to extract it
 * @param [includeExamples=false] - Optional flag, if true, will add the `examplesId` into the list
 * @returns - The string containing the list of ids for use in an `aria-describedBy` attribute
 */
declare function ariaDescribedByIds<T = any>(id: IdSchema<T> | string, includeExamples?: boolean): string;
/** Return a consistent `id` for the `optionIndex`s of a `Radio` or `Checkboxes` widget
 *
 * @param id - The id of the parent component for the option
 * @param optionIndex - The index of the option for which the id is desired
 * @returns - An id for the option index based on the parent `id`
 */
declare function optionId(id: string, optionIndex: number): string;

/** This function checks if the given `schema` matches a single constant value. This happens when either the schema has
 * an `enum` array with a single value or there is a `const` defined.
 *
 * @param schema - The schema for a field
 * @returns - True if the `schema` has a single constant value, false otherwise
 */
declare function isConstant<S extends StrictRJSFSchema = RJSFSchema>(schema: S): boolean;

/** Checks to see if the `uiSchema` contains the `widget` field and that the widget is not `hidden`
 *
 * @param uiSchema - The UI Schema from which to detect if it is customized
 * @returns - True if the `uiSchema` describes a custom widget, false otherwise
 */
declare function isCustomWidget<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any>(uiSchema?: UiSchema<T, S, F>): boolean;

/** Detects whether the given `schema` contains fixed items. This is the case when `schema.items` is a non-empty array
 * that only contains objects.
 *
 * @param schema - The schema in which to check for fixed items
 * @returns - True if there are fixed items in the schema, false otherwise
 */
declare function isFixedItems<S extends StrictRJSFSchema = RJSFSchema>(schema: S): boolean;

/** Determines whether a `thing` is an object for the purposes of RSJF. In this case, `thing` is an object if it has
 * the type `object` but is NOT null, an array or a File.
 *
 * @param thing - The thing to check to see whether it is an object
 * @returns - True if it is a non-null, non-array, non-File object
 */
declare function isObject(thing: any): boolean;

/** Helper function that will return the value to use for a widget `label` based on `hideLabel`. The `fallback` is used
 * as the return value from the function when `hideLabel` is true. Due to the implementation of theme components, it
 * may be necessary to return something other than `undefined` to cause the theme component to not render a label. Some
 * themes require may `false` and others may require an empty string.
 *
 * @param [label] - The label string or component to render when not hidden
 * @param [hideLabel] - Flag, if true, will cause the label to be hidden
 * @param [fallback] - One of 3 values, `undefined` (the default), `false` or an empty string
 * @returns - `fallback` if `hideLabel` is true, otherwise `label`
 */
declare function labelValue(label?: string | ReactElement, hideLabel?: boolean, fallback?: ''): undefined | string;
declare function labelValue(label?: string | ReactElement, hideLabel?: boolean, fallback?: false | ''): string | false | ReactElement<any, string | react.JSXElementConstructor<any>> | undefined;

/** Converts a local Date string into a UTC date string
 *
 * @param dateString - The string representation of a date as accepted by the `Date()` constructor
 * @returns - A UTC date string if `dateString` is truthy, otherwise undefined
 */
declare function localToUTC(dateString: string): string | undefined;

/** Merges the `defaults` object of type `T` into the `formData` of type `T`
 *
 * When merging defaults and form data, we want to merge in this specific way:
 * - objects are deeply merged
 * - arrays are merged in such a way that:
 *   - when the array is set in form data, only array entries set in form data
 *     are deeply merged; additional entries from the defaults are ignored unless `mergeExtraArrayDefaults` is true, in
 *     which case the extras are appended onto the end of the form data
 *   - when the array is not set in form data, the default is copied over
 * - scalars are overwritten/set by form data
 *
 * @param [defaults] - The defaults to merge
 * @param [formData] - The form data into which the defaults will be merged
 * @param [mergeExtraArrayDefaults=false] - If true, any additional default array entries are appended onto the formData
 * @returns - The resulting merged form data with defaults
 */
declare function mergeDefaultsWithFormData<T = any>(defaults?: T, formData?: T, mergeExtraArrayDefaults?: boolean): T | undefined;

/** Recursively merge deeply nested objects.
 *
 * @param obj1 - The first object to merge
 * @param obj2 - The second object to merge
 * @param [concatArrays=false] - Optional flag that, when true, will cause arrays to be concatenated. Use
 *          "preventDuplicates" to merge arrays in a manner that prevents any duplicate entries from being merged.
 *          NOTE: Uses shallow comparison for the duplicate checking.
 * @returns - A new object that is the merge of the two given objects
 */
declare function mergeObjects(obj1: GenericObjectType, obj2: GenericObjectType, concatArrays?: boolean | 'preventDuplicates'): GenericObjectType;

/** Recursively merge deeply nested schemas. The difference between `mergeSchemas` and `mergeObjects` is that
 * `mergeSchemas` only concats arrays for values under the 'required' keyword, and when it does, it doesn't include
 * duplicate values.
 *
 * @param obj1 - The first schema object to merge
 * @param obj2 - The second schema object to merge
 * @returns - The merged schema object
 */
declare function mergeSchemas(obj1: GenericObjectType, obj2: GenericObjectType): GenericObjectType;

/** Gets the list of options from the schema. If the schema has an enum list, then those enum values are returned. The
 * labels for the options will be extracted from the non-standard, RJSF-deprecated `enumNames` if it exists, otherwise
 * the label will be the same as the `value`. If the schema has a `oneOf` or `anyOf`, then the value is the list of
 * `const` values from the schema and the label is either the `schema.title` or the value.
 *
 * @param schema - The schema from which to extract the options list
 * @returns - The list of options from the schema
 */
declare function optionsList<S extends StrictRJSFSchema = RJSFSchema>(schema: S): EnumOptionsType<S>[] | undefined;

/** Given a list of `properties` and an `order` list, returns a list that contains the `properties` ordered correctly.
 * If `order` is not an array, then the untouched `properties` list is returned. Otherwise `properties` is ordered per
 * the `order` list. If `order` contains a '*' then any `properties` that are not mentioned explicity in `order` will be
 * places in the location of the `*`.
 *
 * @param properties - The list of property keys to be ordered
 * @param order - An array of property keys to be ordered first, with an optional '*' property
 * @returns - A list with the `properties` ordered
 * @throws - Error when the properties cannot be ordered correctly
 */
declare function orderProperties(properties: string[], order?: string[]): string[];

/** Returns a string representation of the `num` that is padded with leading "0"s if necessary
 *
 * @param num - The number to pad
 * @param width - The width of the string at which no lead padding is necessary
 * @returns - The number converted to a string with leading zero padding if the number of digits is less than `width`
 */
declare function pad(num: number, width: number): string;

/** Parses the `dateString` into a `DateObject`, including the time information when `includeTime` is true
 *
 * @param dateString - The date string to parse into a DateObject
 * @param [includeTime=true] - Optional flag, if false, will not include the time data into the object
 * @returns - The date string converted to a `DateObject`
 * @throws - Error when the date cannot be parsed from the string
 */
declare function parseDateString(dateString?: string, includeTime?: boolean): DateObject;

/** Extracts the range spec information `{ step?: number, min?: number, max?: number }` that can be spread onto an HTML
 * input from the range analog in the schema `{ multipleOf?: number, minimum?: number, maximum?: number }`.
 *
 * @param schema - The schema from which to extract the range spec
 * @returns - A range specification from the schema
 */
declare function rangeSpec<S extends StrictRJSFSchema = RJSFSchema>(schema: S): RangeSpecType;

/** Potentially substitutes all replaceable parameters with the associated value(s) from the `params` if available. When
 * a `params` array is provided, each value in the array is used to replace any of the replaceable parameters in the
 * `inputString` using the `%1`, `%2`, etc. replacement specifiers.
 *
 * @param inputString - The string which will be potentially updated with replacement parameters
 * @param params - The optional list of replaceable parameter values to substitute into the english string
 * @returns - The updated string with any replacement specifiers replaced
 */
declare function replaceStringParameters(inputString: string, params?: string[]): string;

/** Check to see if a `schema` specifies that a value must be true. This happens when:
 * - `schema.const` is truthy
 * - `schema.enum` == `[true]`
 * - `schema.anyOf` or `schema.oneOf` has a single value which recursively returns true
 * - `schema.allOf` has at least one value which recursively returns true
 *
 * @param schema - The schema to check
 * @returns - True if the schema specifies a value that must be true, false otherwise
 */
declare function schemaRequiresTrueValue<S extends StrictRJSFSchema = RJSFSchema>(schema: S): boolean;

/** Determines whether the given `component` should be rerendered by comparing its current set of props and state
 * against the next set. If either of those two sets are not the same, then the component should be rerendered.
 *
 * @param component - A React component being checked
 * @param nextProps - The next set of props against which to check
 * @param nextState - The next set of state against which to check
 * @returns - True if the component should be re-rendered, false otherwise
 */
declare function shouldRender(component: react__default.Component, nextProps: any, nextState: any): boolean;

/** Returns the constant value from the schema when it is either a single value enum or has a const key. Otherwise
 * throws an error.
 *
 * @param schema - The schema from which to obtain the constant value
 * @returns - The constant value for the schema
 * @throws - Error when the schema does not have a constant value
 */
declare function toConstant<S extends StrictRJSFSchema = RJSFSchema>(schema: S): json_schema.JSONSchema7Type | undefined;

/** Returns a UTC date string for the given `dateObject`. If `time` is false, then the time portion of the string is
 * removed.
 *
 * @param dateObject - The `DateObject` to convert to a date string
 * @param [time=true] - Optional flag used to remove the time portion of the date string if false
 * @returns - The UTC date string
 */
declare function toDateString(dateObject: DateObject, time?: boolean): string;

/** Converts an `errorSchema` into a list of `RJSFValidationErrors`
 *
 * @param errorSchema - The `ErrorSchema` instance to convert
 * @param [fieldPath=[]] - The current field path, defaults to [] if not specified
 * @returns - The list of `RJSFValidationErrors` extracted from the `errorSchema`
 */
declare function toErrorList<T = any>(errorSchema?: ErrorSchema<T>, fieldPath?: string[]): RJSFValidationError[];

/** Transforms a rjsf validation errors list:
 * [
 *   {property: '.level1.level2[2].level3', message: 'err a'},
 *   {property: '.level1.level2[2].level3', message: 'err b'},
 *   {property: '.level1.level2[4].level3', message: 'err b'},
 * ]
 * Into an error tree:
 * {
 *   level1: {
 *     level2: {
 *       2: {level3: {errors: ['err a', 'err b']}},
 *       4: {level3: {errors: ['err b']}},
 *     }
 *   }
 * };
 *
 * @param errors - The list of RJSFValidationError objects
 * @returns - The `ErrorSchema` built from the list of `RJSFValidationErrors`
 */
declare function toErrorSchema<T = any>(errors: RJSFValidationError[]): ErrorSchema<T>;

/** Unwraps the `errorHandler` structure into the associated `ErrorSchema`, stripping the `addError()` functions from it
 *
 * @param errorHandler - The `FormValidation` error handling structure
 * @returns - The `ErrorSchema` resulting from the stripping of the `addError()` function
 */
declare function unwrapErrorHandler<T = any>(errorHandler: FormValidation<T>): ErrorSchema<T>;

/** Converts a UTC date string into a local Date format
 *
 * @param jsonDate - A UTC date string
 * @returns - An empty string when `jsonDate` is falsey, otherwise a date string in local format
 */
declare function utcToLocal(jsonDate: string): string;

/** Merges the errors in `additionalErrorSchema` into the existing `validationData` by combining the hierarchies in the
 * two `ErrorSchema`s and then appending the error list from the `additionalErrorSchema` obtained by calling
 * `toErrorList()` on the `errors` in the `validationData`. If no `additionalErrorSchema` is passed, then
 * `validationData` is returned.
 *
 * @param validationData - The current `ValidationData` into which to merge the additional errors
 * @param [additionalErrorSchema] - The optional additional set of errors in an `ErrorSchema`
 * @returns - The `validationData` with the additional errors from `additionalErrorSchema` merged into it, if provided.
 */
declare function validationDataMerge<T = any>(validationData: ValidationData<T>, additionalErrorSchema?: ErrorSchema<T>): ValidationData<T>;

/** Recursively prefixes all `$ref`s in a schema with the value of the `ROOT_SCHEMA_PREFIX` constant.
 * This is used in isValid to make references to the rootSchema
 *
 * @param schemaNode - The object node to which a ROOT_SCHEMA_PREFIX is added when a REF_KEY is part of it
 * @returns - A copy of the `schemaNode` with updated `$ref`s
 */
declare function withIdRefPrefix<S extends StrictRJSFSchema = RJSFSchema>(schemaNode: S): S | S[];

/** Below are the list of all the keys into various elements of a RJSFSchema or UiSchema that are used by the various
 * utility functions. In addition to those keys, there are the special `ADDITIONAL_PROPERTY_FLAG` and
 * `RJSF_ADDITONAL_PROPERTIES_FLAG` flags that is added to a schema under certain conditions by the `retrieveSchema()`
 * utility.
 */
declare const ADDITIONAL_PROPERTY_FLAG = "__additional_property";
declare const ADDITIONAL_PROPERTIES_KEY = "additionalProperties";
declare const ALL_OF_KEY = "allOf";
declare const ANY_OF_KEY = "anyOf";
declare const CONST_KEY = "const";
declare const DEFAULT_KEY = "default";
declare const DEFINITIONS_KEY = "definitions";
declare const DEPENDENCIES_KEY = "dependencies";
declare const ENUM_KEY = "enum";
declare const ERRORS_KEY = "__errors";
declare const ID_KEY = "$id";
declare const IF_KEY = "if";
declare const ITEMS_KEY = "items";
declare const JUNK_OPTION_ID = "_$junk_option_schema_id$_";
declare const NAME_KEY = "$name";
declare const ONE_OF_KEY = "oneOf";
declare const PROPERTIES_KEY = "properties";
declare const REQUIRED_KEY = "required";
declare const SUBMIT_BTN_OPTIONS_KEY = "submitButtonOptions";
declare const REF_KEY = "$ref";
declare const RJSF_ADDITONAL_PROPERTIES_FLAG = "__rjsf_additionalProperties";
declare const ROOT_SCHEMA_PREFIX = "__rjsf_rootSchema";
declare const UI_FIELD_KEY = "ui:field";
declare const UI_WIDGET_KEY = "ui:widget";
declare const UI_OPTIONS_KEY = "ui:options";
declare const UI_GLOBAL_OPTIONS_KEY = "ui:globalOptions";

/** The type of the map of schema hash to schema
 */
type SchemaMap<S extends StrictRJSFSchema = RJSFSchema> = {
    [hash: string]: S;
};

/** Parses the given `rootSchema` to extract out all the sub-schemas that maybe contained within it. Returns a map of
 * the hash of the schema to schema/sub-schema.
 *
 * @param rootSchema - The root schema to parse for sub-schemas used by `isValid()` calls
 * @returns - The `SchemaMap` of all schemas that were parsed
 */
declare function schemaParser<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any>(rootSchema: S): SchemaMap<S>;

/** Returns the superset of `formData` that includes the given set updated to include any missing fields that have
 * computed to have defaults provided in the `schema`.
 *
 * @param validator - An implementation of the `ValidatorType` interface that will be used when necessary
 * @param theSchema - The schema for which the default state is desired
 * @param [formData] - The current formData, if any, onto which to provide any missing defaults
 * @param [rootSchema] - The root schema, used to primarily to look up `$ref`s
 * @param [includeUndefinedValues=false] - Optional flag, if true, cause undefined values to be added as defaults.
 *          If "excludeObjectChildren", cause undefined values for this object and pass `includeUndefinedValues` as
 *          false when computing defaults for any nested object properties.
 * @param [experimental_defaultFormStateBehavior] Optional configuration object, if provided, allows users to override default form state behavior
 * @returns - The resulting `formData` with all the defaults provided
 */
declare function getDefaultFormState<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any>(validator: ValidatorType<T, S, F>, theSchema: S, formData?: T, rootSchema?: S, includeUndefinedValues?: boolean | 'excludeObjectChildren', experimental_defaultFormStateBehavior?: Experimental_DefaultFormStateBehavior): T | T[] | undefined;

/** Determines whether the combination of `schema` and `uiSchema` properties indicates that the label for the `schema`
 * should be displayed in a UI.
 *
 * @param validator - An implementation of the `ValidatorType` interface that will be used when necessary
 * @param schema - The schema for which the display label flag is desired
 * @param [uiSchema={}] - The UI schema from which to derive potentially displayable information
 * @param [rootSchema] - The root schema, used to primarily to look up `$ref`s
 * @param [globalOptions={}] - The optional Global UI Schema from which to get any fallback `xxx` options
 * @returns - True if the label should be displayed or false if it should not
 */
declare function getDisplayLabel<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any>(validator: ValidatorType<T, S, F>, schema: S, uiSchema?: UiSchema<T, S, F>, rootSchema?: S, globalOptions?: GlobalUISchemaOptions): boolean;

/** Determines which of the given `options` provided most closely matches the `formData`. Using
 * `getFirstMatchingOption()` to match two schemas that differ only by the readOnly, default or const value of a field
 * based on the `formData` and returns 0 when there is no match. Rather than passing in all the `options` at once to
 * this utility, instead an array of valid option indexes is created by iterating over the list of options, call
 * `getFirstMatchingOptions` with a list of one junk option and one good option, seeing if the good option is considered
 * matched.
 *
 * Once the list of valid indexes is created, if there is only one valid index, just return it. Otherwise, if there are
 * no valid indexes, then fill the valid indexes array with the indexes of all the options. Next, the index of the
 * option with the highest score is determined by iterating over the list of valid options, calling
 * `calculateIndexScore()` on each, comparing it against the current best score, and returning the index of the one that
 * eventually has the best score.
 *
 * @param validator - An implementation of the `ValidatorType` interface that will be used when necessary
 * @param rootSchema - The root JSON schema of the entire form
 * @param formData - The form data associated with the schema
 * @param options - The list of options that can be selected from
 * @param [selectedOption=-1] - The index of the currently selected option, defaulted to -1 if not specified
 * @param [discriminatorField] - The optional name of the field within the options object whose value is used to
 *          determine which option is selected
 * @returns - The index of the option that is the closest match to the `formData` or the `selectedOption` if no match
 */
declare function getClosestMatchingOption<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any>(validator: ValidatorType<T, S, F>, rootSchema: S, formData: T | undefined, options: S[], selectedOption?: number, discriminatorField?: string): number;

/** Given the `formData` and list of `options`, attempts to find the index of the first option that matches the data.
 * Always returns the first option if there is nothing that matches.
 *
 * @param validator - An implementation of the `ValidatorType` interface that will be used when necessary
 * @param formData - The current formData, if any, used to figure out a match
 * @param options - The list of options to find a matching options from
 * @param rootSchema - The root schema, used to primarily to look up `$ref`s
 * @param [discriminatorField] - The optional name of the field within the options object whose value is used to
 *          determine which option is selected
 * @returns - The index of the first matched option or 0 if none is available
 */
declare function getFirstMatchingOption<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any>(validator: ValidatorType<T, S, F>, formData: T | undefined, options: S[], rootSchema: S, discriminatorField?: string): number;

/** Given the `formData` and list of `options`, attempts to find the index of the option that best matches the data.
 * Deprecated, use `getFirstMatchingOption()` instead.
 *
 * @param validator - An implementation of the `ValidatorType` interface that will be used when necessary
 * @param formData - The current formData, if any, used to figure out a match
 * @param options - The list of options to find a matching options from
 * @param rootSchema - The root schema, used to primarily to look up `$ref`s
 * @param [discriminatorField] - The optional name of the field within the options object whose value is used to
 *          determine which option is selected
 * @returns - The index of the matched option or 0 if none is available
 * @deprecated
 */
declare function getMatchingOption<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any>(validator: ValidatorType<T, S, F>, formData: T | undefined, options: S[], rootSchema: S, discriminatorField?: string): number;

/** Checks to see if the `schema` and `uiSchema` combination represents an array of files
 *
 * @param validator - An implementation of the `ValidatorType` interface that will be used when necessary
 * @param schema - The schema for which check for array of files flag is desired
 * @param [uiSchema={}] - The UI schema from which to check the widget
 * @param [rootSchema] - The root schema, used to primarily to look up `$ref`s
 * @returns - True if schema/uiSchema contains an array of files, otherwise false
 */
declare function isFilesArray<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any>(validator: ValidatorType<T, S, F>, schema: S, uiSchema?: UiSchema<T, S, F>, rootSchema?: S): boolean;

/** Checks to see if the `schema` combination represents a multi-select
 *
 * @param validator - An implementation of the `ValidatorType` interface that will be used when necessary
 * @param schema - The schema for which check for a multi-select flag is desired
 * @param [rootSchema] - The root schema, used to primarily to look up `$ref`s
 * @returns - True if schema contains a multi-select, otherwise false
 */
declare function isMultiSelect<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any>(validator: ValidatorType<T, S, F>, schema: S, rootSchema?: S): boolean;

/** Checks to see if the `schema` combination represents a select
 *
 * @param validator - An implementation of the `ValidatorType` interface that will be used when necessary
 * @param theSchema - The schema for which check for a select flag is desired
 * @param [rootSchema] - The root schema, used to primarily to look up `$ref`s
 * @returns - True if schema contains a select, otherwise false
 */
declare function isSelect<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any>(validator: ValidatorType<T, S, F>, theSchema: S, rootSchema?: S): boolean;

/** Merges the errors in `additionalErrorSchema` into the existing `validationData` by combining the hierarchies in the
 * two `ErrorSchema`s and then appending the error list from the `additionalErrorSchema` obtained by calling
 * `validator.toErrorList()` onto the `errors` in the `validationData`. If no `additionalErrorSchema` is passed, then
 * `validationData` is returned.
 *
 * @param validator - The validator used to convert an ErrorSchema to a list of errors
 * @param validationData - The current `ValidationData` into which to merge the additional errors
 * @param [additionalErrorSchema] - The additional set of errors in an `ErrorSchema`
 * @returns - The `validationData` with the additional errors from `additionalErrorSchema` merged into it, if provided.
 * @deprecated - Use the `validationDataMerge()` function exported from `@rjsf/utils` instead. This function will be
 *        removed in the next major release.
 */
declare function mergeValidationData<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any>(validator: ValidatorType<T, S, F>, validationData: ValidationData<T>, additionalErrorSchema?: ErrorSchema<T>): ValidationData<T>;

/** Retrieves an expanded schema that has had all of its conditions, additional properties, references and dependencies
 * resolved and merged into the `schema` given a `validator`, `rootSchema` and `rawFormData` that is used to do the
 * potentially recursive resolution.
 *
 * @param validator - An implementation of the `ValidatorType` interface that will be forwarded to all the APIs
 * @param schema - The schema for which retrieving a schema is desired
 * @param [rootSchema={}] - The root schema that will be forwarded to all the APIs
 * @param [rawFormData] - The current formData, if any, to assist retrieving a schema
 * @returns - The schema having its conditions, additional properties, references and dependencies resolved
 */
declare function retrieveSchema<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any>(validator: ValidatorType<T, S, F>, schema: S, rootSchema?: S, rawFormData?: T): S;

/** Sanitize the `data` associated with the `oldSchema` so it is considered appropriate for the `newSchema`. If the new
 * schema does not contain any properties, then `undefined` is returned to clear all the form data. Due to the nature
 * of schemas, this sanitization happens recursively for nested objects of data. Also, any properties in the old schema
 * that are non-existent in the new schema are set to `undefined`. The data sanitization process has the following flow:
 *
 * - If the new schema is an object that contains a `properties` object then:
 *   - Create a `removeOldSchemaData` object, setting each key in the `oldSchema.properties` having `data` to undefined
 *   - Create an empty `nestedData` object for use in the key filtering below:
 *   - Iterate over each key in the `newSchema.properties` as follows:
 *     - Get the `formValue` of the key from the `data`
 *     - Get the `oldKeySchema` and `newKeyedSchema` for the key, defaulting to `{}` when it doesn't exist
 *     - Retrieve the schema for any refs within each `oldKeySchema` and/or `newKeySchema`
 *     - Get the types of the old and new keyed schemas and if the old doesn't exist or the old & new are the same then:
 *       - If `removeOldSchemaData` has an entry for the key, delete it since the new schema has the same property
 *       - If type of the key in the new schema is `object`:
 *         - Store the value from the recursive `sanitizeDataForNewSchema` call in `nestedData[key]`
 *       - Otherwise, check for default or const values:
 *         - Get the old and new `default` values from the schema and check:
 *           - If the new `default` value does not match the form value:
 *             - If the old `default` value DOES match the form value, then:
 *               - Replace `removeOldSchemaData[key]` with the new `default`
 *               - Otherwise, if the new schema is `readOnly` then replace `removeOldSchemaData[key]` with undefined
 *         - Get the old and new `const` values from the schema and check:
 *           - If the new `const` value does not match the form value:
 *           - If the old `const` value DOES match the form value, then:
 *             - Replace `removeOldSchemaData[key]` with the new `const`
 *             - Otherwise, replace `removeOldSchemaData[key]` with undefined
 *   - Once all keys have been processed, return an object built as follows:
 *     - `{ ...removeOldSchemaData, ...nestedData, ...pick(data, keysToKeep) }`
 * - If the new and old schema types are array and the `data` is an array then:
 *   - If the type of the old and new schema `items` are a non-array objects:
 *     - Retrieve the schema for any refs within each `oldKeySchema.items` and/or `newKeySchema.items`
 *     - If the `type`s of both items are the same (or the old does not have a type):
 *       - If the type is "object", then:
 *         - For each element in the `data` recursively sanitize the data, stopping at `maxItems` if specified
 *       - Otherwise, just return the `data` removing any values after `maxItems` if it is set
 *   - If the type of the old and new schema `items` are booleans of the same value, return `data` as is
 * - Otherwise return `undefined`
 *
 * @param validator - An implementation of the `ValidatorType` interface that will be used when necessary
 * @param rootSchema - The root JSON schema of the entire form
 * @param [newSchema] - The new schema for which the data is being sanitized
 * @param [oldSchema] - The old schema from which the data originated
 * @param [data={}] - The form data associated with the schema, defaulting to an empty object when undefined
 * @returns - The new form data, with all the fields uniquely associated with the old schema set
 *      to `undefined`. Will return `undefined` if the new schema is not an object containing properties.
 */
declare function sanitizeDataForNewSchema<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any>(validator: ValidatorType<T, S, F>, rootSchema: S, newSchema?: S, oldSchema?: S, data?: any): T;

/** Generates an `IdSchema` object for the `schema`, recursively
 *
 * @param validator - An implementation of the `ValidatorType` interface that will be used when necessary
 * @param schema - The schema for which the `IdSchema` is desired
 * @param [id] - The base id for the schema
 * @param [rootSchema] - The root schema, used to primarily to look up `$ref`s
 * @param [formData] - The current formData, if any, to assist retrieving a schema
 * @param [idPrefix='root'] - The prefix to use for the id
 * @param [idSeparator='_'] - The separator to use for the path segments in the id
 * @returns - The `IdSchema` object for the `schema`
 */
declare function toIdSchema<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any>(validator: ValidatorType<T, S, F>, schema: S, id?: string | null, rootSchema?: S, formData?: T, idPrefix?: string, idSeparator?: string): IdSchema<T>;

/** Generates an `PathSchema` object for the `schema`, recursively
 *
 * @param validator - An implementation of the `ValidatorType` interface that will be used when necessary
 * @param schema - The schema for which the `PathSchema` is desired
 * @param [name=''] - The base name for the schema
 * @param [rootSchema] - The root schema, used to primarily to look up `$ref`s
 * @param [formData] - The current formData, if any, to assist retrieving a schema
 * @returns - The `PathSchema` object for the `schema`
 */
declare function toPathSchema<T = any, S extends StrictRJSFSchema = RJSFSchema, F extends FormContextType = any>(validator: ValidatorType<T, S, F>, schema: S, name?: string, rootSchema?: S, formData?: T): PathSchema<T>;

export { ADDITIONAL_PROPERTIES_KEY, ADDITIONAL_PROPERTY_FLAG, ALL_OF_KEY, ANY_OF_KEY, ArrayFieldDescriptionProps, ArrayFieldTemplateItemType, ArrayFieldTemplateProps, ArrayFieldTitleProps, BaseInputTemplateProps, CONST_KEY, CustomValidator, DEFAULT_KEY, DEFINITIONS_KEY, DEPENDENCIES_KEY, DateObject, DescriptionFieldProps, ENUM_KEY, ERRORS_KEY, EnumOptionsType, ErrorListProps, ErrorSchema, ErrorSchemaBuilder, ErrorTransformer, Experimental_ArrayMinItems, Experimental_DefaultFormStateBehavior, Field, FieldError, FieldErrorProps, FieldErrors, FieldHelpProps, FieldId, FieldPath, FieldProps, FieldTemplateProps, FieldValidation, FormContextType, FormValidation, GenericObjectType, GlobalUISchemaOptions, ID_KEY, IF_KEY, ITEMS_KEY, IconButtonProps, IdSchema, InputPropsType, JUNK_OPTION_ID, NAME_KEY, ONE_OF_KEY, ObjectFieldTemplatePropertyType, ObjectFieldTemplateProps, PROPERTIES_KEY, PathSchema, REF_KEY, REQUIRED_KEY, RJSFSchema, RJSFValidationError, RJSF_ADDITONAL_PROPERTIES_FLAG, ROOT_SCHEMA_PREFIX, RangeSpecType, Registry, RegistryFieldsType, RegistryWidgetsType, SUBMIT_BTN_OPTIONS_KEY, SchemaMap, SchemaUtilsType, StrictRJSFSchema, SubmitButtonProps, TemplatesType, TitleFieldProps, TranslatableString, UIOptionsType, UISchemaSubmitButtonOptions, UI_FIELD_KEY, UI_GLOBAL_OPTIONS_KEY, UI_OPTIONS_KEY, UI_WIDGET_KEY, UiSchema, UnsupportedFieldProps, ValidationData, ValidatorType, Widget, WidgetProps, WrapIfAdditionalTemplateProps, allowAdditionalItems, ariaDescribedByIds, asNumber, canExpand, createErrorHandler, createSchemaUtils, dataURItoBlob, deepEquals, descriptionId, englishStringTranslator, enumOptionsDeselectValue, enumOptionsIndexForValue, enumOptionsIsSelected, enumOptionsSelectValue, enumOptionsValueForIndex, errorId, examplesId, findSchemaDefinition, getClosestMatchingOption, getDefaultFormState, getDiscriminatorFieldFromSchema, getDisplayLabel, getFirstMatchingOption, getInputProps, getMatchingOption, getSchemaType, getSubmitButtonOptions, getTemplate, getUiOptions, getWidget, guessType, hasWidget, hashForSchema, helpId, isConstant, isCustomWidget, isFilesArray, isFixedItems, isMultiSelect, isObject, isSelect, labelValue, localToUTC, mergeDefaultsWithFormData, mergeObjects, mergeSchemas, mergeValidationData, optionId, optionsList, orderProperties, pad, parseDateString, rangeSpec, replaceStringParameters, retrieveSchema, sanitizeDataForNewSchema, schemaParser, schemaRequiresTrueValue, shouldRender, titleId, toConstant, toDateString, toErrorList, toErrorSchema, toIdSchema, toPathSchema, unwrapErrorHandler, utcToLocal, validationDataMerge, withIdRefPrefix };
